[llvm] e80ee4c - [GlobalDCE] In VFE support for relative pointers, allow GEP references to the base symbol

Kuba Mracek via llvm-commits llvm-commits at lists.llvm.org
Mon Sep 13 15:22:26 PDT 2021


Author: Kuba Mracek
Date: 2021-09-13T15:22:11-07:00
New Revision: e80ee4cbd9a79efa99cabdbb42faba7c7b01e70e

URL: https://github.com/llvm/llvm-project/commit/e80ee4cbd9a79efa99cabdbb42faba7c7b01e70e
DIFF: https://github.com/llvm/llvm-project/commit/e80ee4cbd9a79efa99cabdbb42faba7c7b01e70e.diff

LOG: [GlobalDCE] In VFE support for relative pointers, allow GEP references to the base symbol

This is for Swift VFE support. In some vtable forms that Swift emits, the "base" of a relative pointer is not the global symbol itself directly, but a GEP into it -- so the pointer is relative to a particular field in the global. So getPointerAtOffset() needs to be able to see through the GEP and allow it in a SUB expression, to correctly recognize the offset as a vtable slot.

Differential Revision: https://reviews.llvm.org/D109169

Added: 
    llvm/test/Transforms/GlobalDCE/virtual-functions-relative-pointers-gep.ll

Modified: 
    llvm/lib/Analysis/TypeMetadataUtils.cpp

Removed: 
    


################################################################################
diff  --git a/llvm/test/Transforms/GlobalDCE/virtual-functions-relative-pointers-gep.ll b/llvm/test/Transforms/GlobalDCE/virtual-functions-relative-pointers-gep.ll
new file mode 100644
index 0000000000000..e49dc784af612
--- /dev/null
+++ b/llvm/test/Transforms/GlobalDCE/virtual-functions-relative-pointers-gep.ll
@@ -0,0 +1,41 @@
+; RUN: opt < %s -passes='globaldce' -S | FileCheck %s
+
+target datalayout = "e-m:e-i64:64-f80:128-n8:16:32:64-S128"
+
+declare { i8*, i1 } @llvm.type.checked.load(i8*, i32, metadata)
+
+; A vtable with "relative pointers", slots don't contain pointers to implementations, but instead have an i32 offset from the vtable itself to the implementation.
+ at vtable = internal unnamed_addr constant { [4 x i32] } { [4 x i32] [
+  i32 42,
+  i32 1337,
+  i32 trunc (i64 sub (i64 ptrtoint (void ()* @vfunc1_live              to i64), i64 ptrtoint (i32* getelementptr inbounds ({ [4 x i32] }, { [4 x i32] }* @vtable, i32 0, i32 0, i32 2) to i64)) to i32),
+  i32 trunc (i64 sub (i64 ptrtoint (void ()* @vfunc2_dead              to i64), i64 ptrtoint (i32* getelementptr inbounds ({ [4 x i32] }, { [4 x i32] }* @vtable, i32 0, i32 0, i32 2) to i64)) to i32)
+]}, align 8, !type !0, !type !1, !vcall_visibility !{i64 2}
+!0 = !{i64 8, !"vfunc1.type"}
+!1 = !{i64 12, !"vfunc2.type"}
+
+; CHECK:      @vtable = internal unnamed_addr constant { [4 x i32] } { [4 x i32] [
+; CHECK-SAME:   i32 trunc (i64 sub (i64 ptrtoint (void ()* @vfunc1_live              to i64), i64 ptrtoint (i32* getelementptr inbounds ({ [4 x i32] }, { [4 x i32] }* @vtable, i32 0, i32 0, i32 2) to i64)) to i32),
+; CHECK-SAME:   i32 trunc (i64 sub (i64 0,                                                    i64 ptrtoint (i32* getelementptr inbounds ({ [4 x i32] }, { [4 x i32] }* @vtable, i32 0, i32 0, i32 2) to i64)) to i32)
+; CHECK-SAME: ] }, align 8, !type !0, !type !1, !vcall_visibility !2
+
+; (1) vfunc1_live is referenced from @main, stays alive
+define internal void @vfunc1_live() {
+  ; CHECK: define internal void @vfunc1_live(
+  ret void
+}
+
+; (2) vfunc2_dead is never referenced, gets removed and vtable slot is null'd
+define internal void @vfunc2_dead() {
+  ; CHECK-NOT: define internal void @vfunc2_dead(
+  ret void
+}
+
+define void @main() {
+  %1 = ptrtoint { [4 x i32] }* @vtable to i64 ; to keep @vtable alive
+  %2 = tail call { i8*, i1 } @llvm.type.checked.load(i8* null, i32 0, metadata !"vfunc1.type")
+  ret void
+}
+
+!999 = !{i32 1, !"Virtual Function Elim", i32 1}
+!llvm.module.flags = !{!999}


        


More information about the llvm-commits mailing list
diff  --git a/llvm/lib/Analysis/TypeMetadataUtils.cpp b/llvm/lib/Analysis/TypeMetadataUtils.cpp
index c55071a3685f4..65fa2bbcffd3c 100644
--- a/llvm/lib/Analysis/TypeMetadataUtils.cpp
+++ b/llvm/lib/Analysis/TypeMetadataUtils.cpp
@@ -173,10 +173,19 @@ Constant *llvm::getPointerAtOffset(Constant *I, uint64_t Offset, Module &M,
     case Instruction::Sub: {
       auto *Operand0 = cast<Constant>(C->getOperand(0));
       auto *Operand1 = cast<Constant>(C->getOperand(1));
-      auto *Operand1TargetGlobal = getPointerAtOffset(Operand1, 0, M);
+
+      auto StripGEP = [](Constant *C) {
+        auto *CE = dyn_cast<ConstantExpr>(C);
+        if (!CE)
+          return C;
+        if (CE->getOpcode() != Instruction::GetElementPtr)
+          return C;
+        return CE->getOperand(0);
+      };
+      auto *Operand1TargetGlobal = StripGEP(getPointerAtOffset(Operand1, 0, M));
 
       // Check that in the "sub (@a, @b)" expression, @b points back to the top
       // level global that we're processing. Otherwise bail.
       if (Operand1TargetGlobal != TopLevelGlobal)
         return nullptr;
 
@@ -173,10 +173,19 @@ Constant *llvm::getPointerAtOffset(Constant *I, uint64_t Offset, Module &M,
 
       // Check that in the "sub (@a, @b)" expression, @b points back to the top
-      // level global that we're processing. Otherwise bail.
+      // level global (or a GEP thereof) that we're processing. Otherwise bail.
       if (Operand1TargetGlobal != TopLevelGlobal)
         return nullptr;
